from __future__ import annotations
from robot.xml_engine.pyquery_engine import PyQueryAdapter
from robot.http_engine.aiohttp_engine import AioHttpAdapter
from robot.api import Context, XmlEngine, HttpEngine, HttpSession, XmlNode
from dataclasses import dataclass, field
from typing import Tuple, Any, Dict
from urllib.parse import urlparse

import logging
from logging import Logger

__logger__ = logging.getLogger(__name__)


@dataclass()
class ContextImpl(Context):
    xml_engine: XmlEngine = field(default_factory=PyQueryAdapter)
    http_engine: HttpEngine = field(default_factory=AioHttpAdapter)
    url: Any = None
    http_headers: Dict[str, str] = field(default_factory=dict)
    http_session: HttpSession = None
    logger: Logger = field(default=__logger__)

    def __iter__(self):
        yield 'url', self.url.geturl()
        yield 'headers', dict(self.http_headers)

    def clone(self, **override) -> ContextImpl:
        new_instance = self.__class__(**self.__dict__)
        for k, v in override.items():
            setattr(new_instance, k, v)
        return new_instance

    async def close(self):
        if self.http_session is None:
            return
        await self.http_session.close()
        self.http_session = None

    def get_http_session(self) -> HttpSession:
        if self.http_session is not None:
            return self.http_session
        self.http_session = self.http_engine.session()
        return self.http_session

    def resolve_url(self, url: str):
        if self.url is None:
            return url
        if '://' in url:
            return url
        if url.startswith('//'):
            return self.url.scheme + ':' + url
        if url.startswith('/'):
            return self.url.scheme + '://' + self.url.netloc + url
        base_path, _ = self.url.path.rsplit('/', 1)
        return ''.join([
            self.url.scheme,
            '://',
            self.url.netloc,
            base_path,
            '/',
            url,
        ])

    async def download(self, url: str, filename: str):
        session = self.get_http_session()
        await session.download(url, filename)

    async def http_get(self, url) -> Tuple[Context, XmlNode]:
        session = self.get_http_session()
        url = self.resolve_url(url)
        headers, response = await session.get(url)
        return self.clone(
            url=urlparse(url),
            http_headers=headers,
            http_session=session,
        ), self.xml_engine(response)



import csv
import json
import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import Any, Callable, Sequence, Dict, Tuple

from robot.api import Collector, Context

__logger__ = logging.getLogger(__name__)


@dataclass()
class StoreCollector(Collector[Any, str]):
    filename: Collector[Any, str]
    mode: str = 'w+'
    serializer: Callable[[Any], str] = json.dumps
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Any) -> Tuple[Context, str]:
        context, filename = await self.filename(context, item)
        with open(filename, self.mode) as output:
            output.write(self.serializer(item))
        return context, filename


@dataclass()
class CsvCollector(Collector[Sequence[Sequence[Any]], str]):
    filename: Collector[Any, str]
    mode: str = 'w+'
    csv_writer_factory = csv.writer
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Sequence[Sequence[Any]]) -> Tuple[Context, str]:
        context, filename = await self.filename(context, item)
        with open(filename, self.mode) as output:
            csv_writer = self.csv_writer_factory(output)
            csv_writer.writerows(item)
        return context, filename


@dataclass()
class DictCsvCollector(Collector[Sequence[Dict[str, Any]], str]):
    filename: Collector[Any, str]
    fields: Sequence[str] = None
    mode: str = 'w+'
    csv_writer_factory = csv.DictWriter
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Sequence[Dict[str, Any]]) -> Tuple[Context, str]:
        context, filename = await self.filename(context, item)
        fields = self.fields
        with open(filename, self.mode) as output:
            iterable = iter(item)
            first_item = next(iterable)
            if fields is None:
                fields = sorted(first_item.keys())
            csv_writer = self.csv_writer_factory(output, fields)
            csv_writer.writeheader()
            csv_writer.writerow(first_item)
            csv_writer.writerows(iterable)
        return context, filename



import asyncio
import multiprocessing
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass, field

import logging
from logging import Logger
from robot.api import Robot, Collector, Y, Context
from robot.context.core import ContextImpl

__logger__: Logger = logging.getLogger(__name__)


@dataclass()
class RobotImpl(Robot):
    context: Context = field(default_factory=ContextImpl)
    logger: Logger = field(default=__logger__)

    _loop = None
    _thread_poll = None

    def __enter__(self):
        cpu_count = multiprocessing.cpu_count()
        thread_pool = ThreadPoolExecutor(cpu_count)
        self._loop = asyncio.get_event_loop()
        self._loop.set_default_executor(thread_pool)
        self._thread_poll = thread_pool.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._loop.run_until_complete(
            self.context.close()
        )
        self._thread_poll.__exit__(exc_type, exc_val, exc_tb)

    async def run(self, collector: Collector[None, Y]) -> Y:
        _, result = await collector(self.context, None)
        return result

    def sync_run(self, collector: Collector[None, Y]) -> Y:
        return self._loop.run_until_complete(self.run(collector))



import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import TypeVar, Tuple, Any

import aiohttp

from robot.api import HttpSession, HttpEngine

__logger__: Logger = logging.getLogger(__name__)

T = TypeVar('T')


@dataclass()
class AioHttpSessionAdapter(HttpSession):
    client_session: aiohttp.ClientSession
    logger: Logger = field(default=__logger__)

    async def download(self, url: str, filename: str):
        self.logger.debug(f'starting download {url} to {filename}')
        async with self.client_session.get(url, allow_redirects=True) as response:
            if response.status != 200:
                self.logger.warning(f'download {url} fail with status {response.status}')
                raise Exception()
            with open(filename, 'wb') as output:
                output.write(await response.read())

    async def get(self, url) -> Tuple[Any, str]:
        self.logger.debug(f'starting http get {url}')
        async with self.client_session.get(url, allow_redirects=True) as response:
            self.logger.info(f'http get {url} status {response.status}')
            content = await response.content.read()
            return response.headers, content

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        return await self.close()

    async def close(self):
        return await self.client_session.close()


class AioHttpAdapter(HttpEngine):
    def __init__(self, aiohttp=aiohttp):
        self.aiohttp = aiohttp

    def session(self) -> HttpSession:
        return AioHttpSessionAdapter(self.aiohttp.ClientSession())



from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import List, Any, Callable, Iterable, Dict, Tuple, Awaitable, AsyncIterable, Sequence, Union

from robot.api import Collector, Context, X, Y

__logger__ = logging.getLogger(__name__)


@dataclass(init=False)
class PipeCollector(Collector[Any, Any]):
    collectors: Tuple[Collector[Any, Any]]
    logger: Logger = field(default=__logger__, compare=False)

    def __init__(self, *collectors: Collector[Any, Any], logger=__logger__):
        self.collectors = collectors
        self.logger = logger

    async def __call__(self, context: Context, item: Any) -> Tuple[Context, Any]:
        for collector in self.collectors:
            context, item = await collector(context, item)
        return context, item

    @classmethod
    def from_(cls, *collector: Union[Collector, Sequence[Union]]) -> Collector:
        if len(collector) == 0:
            raise Exception()
        if len(collector) > 1:
            return cls(*collector)
        collector = collector[0]
        if isinstance(collector, (tuple, list,)):
            return cls(*collector)
        else:
            return collector


@dataclass(init=False)
class DefaultCollector(Collector[Any, Any]):
    collectors: Tuple[Collector[Any, Any]]
    logger: Logger = field(default=__logger__, compare=False)

    def __init__(self, *collectors: Collector[Any, Any], logger=__logger__):
        self.collectors = collectors
        self.logger = logger

    def is_empty(self, value):
        if value is None:
            return True
        if isinstance(value, (str,)):
            if value.strip() == '':
                return True
            return False
        if isinstance(value, Iterable):
            for item in value:
                if not self.is_empty(item):
                    return False
            return True
        return False

    async def __call__(self, context: Context, item: Any) -> Tuple[Context, Any]:
        for collector in self.collectors:
            context, result = await collector(context, item)
            if not self.is_empty(result):
                return context, result
        return context, None


@dataclass()
class FnCollector(Collector[X, Y]):
    fn: Callable[[X], Y]
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: X) -> Tuple[Context, Y]:
        return context, self.fn(item)


@dataclass()
class AsyncFnCollector(Collector[X, Y]):
    fn: Callable[[X], Awaitable[Y]]
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: X) -> Tuple[Context, Y]:
        return context, await self.fn(item)


@dataclass()
class NoopCollector(Collector[X, X]):
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: X) -> Tuple[Context, X]:
        return context, item


NOOP_COLLECTOR = NoopCollector()


@dataclass()
class ConstCollector(Collector[Any, Y]):
    value: Y
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Any) -> Tuple[Context, Y]:
        return context, self.value


@dataclass(init=False)
class ForeachCollector(Collector[Sequence[X], Sequence[Y]]):
    collector: Collector[X, Y]
    limit: int = None
    logger: Logger = field(default=__logger__, compare=False)

    def __init__(self, *collector: Collector[X, Y], limit: int = None, logger: Logger = __logger__):
        self.collector = PipeCollector.from_(collector)
        self.limit = limit
        self.logger = logger

    async def __call__(self, context: Context, item: Sequence[X]) -> Tuple[Context, Sequence[Y]]:
        all_items = list(i for i in item)
        batch_size = len(all_items) if self.limit is None else self.limit
        collected_items = []
        for batch in [all_items[index:index + batch_size] for index in range(0, len(all_items), batch_size)]:
            batch_collected_items = await asyncio.gather(*[
                self.collector(context, i) for i in batch
            ])
            collected_items += batch_collected_items
        return context, list(map(lambda it: it[1], collected_items))


@dataclass(init=False)
class AsyncForeachCollector(Collector[AsyncIterable[X], Sequence[Y]]):
    collector: Union[Collector[X, Y], Sequence[Collector]]
    logger: Logger = field(default=__logger__, compare=False)

    def __init__(self, *collector: Collector[X, Y], logger: Logger = __logger__):
        self.collector = PipeCollector.from_(collector)
        self.logger = logger

    async def __call__(self, context: Context, item: AsyncIterable[X]) -> Tuple[Context, Sequence[Y]]:
        tasks = []
        async for i in item:
            coro = self.collector(context, i)
            task = asyncio.create_task(coro)
            tasks.append(task)
        values = await asyncio.gather(*tasks)
        return context, list(map(lambda it: it[1], values))


@dataclass(init=False)
class TupleCollector(Collector[X, Tuple]):
    collectors: Tuple[Collector[X, Any]]
    logger: Logger = field(default=__logger__, compare=False)

    def __init__(self, *collectors: Collector[X, Any], logger=__logger__):
        self.collectors = collectors
        self.logger = logger

    async def __call__(self, context: Context, item: X) -> Tuple[Context, Tuple]:
        collected_items = await asyncio.gather(*[
            collector(context, item)
            for collector in self.collectors
        ])
        return context, tuple(map(lambda it: it[1], collected_items))


@dataclass()
class DelayCollector(Collector[X, X]):
    delay: float
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: X) -> Tuple[Context, X]:
        await asyncio.sleep(self.delay)
        return context, item


@dataclass()
class DictCollector(Collector[X, Dict[str, Any]]):
    nested_collectors: Sequence[Collector[X, Dict[str, Any]]] = ()
    field_collectors: Dict[str, Collector[X, Any]] = field(default_factory=dict)
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: X) -> Tuple[Context, Dict[str, Any]]:
        obj = dict()
        collected_items = await asyncio.gather(*[
            collector(context, item)
            for collector in self.nested_collectors
        ])
        for _, collected_item in collected_items:
            obj.update(collected_item)
        if not self.field_collectors:
            return context, obj
        keys, collectors = zip(*self.field_collectors.items())
        values = await asyncio.gather(*[
            collector(context, item)
            for collector in collectors
        ])
        values = map(lambda it: it[1], values)
        obj.update(dict(zip(keys, values)))
        return context, obj


@dataclass()
class ContextCollector(Collector[Any, Dict[str, Any]]):
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Any) -> Tuple[Context, Dict[str, Any]]:
        return context, dict(context)


CONTEXT = ContextCollector()


@dataclass()
class TapCollector(Collector[X, X]):
    collector: Collector[X, Any]
    logger: Logger = field(default=__logger__, compare=False)

    def __post_init__(self):
        self.collector = PipeCollector.from_(self.collector)

    async def __call__(self, context: Context, item: X) -> Tuple[Context, X]:
        await self.collector(context, item)
        return context, item



from typing import Any, Dict

from robot.api import Collector, X
from robot.collector import collections as _collections
from robot.collector import core as _core
from robot.collector import object as _object
# noinspection PyUnresolvedReferences
from robot.collector.collections import AnyCollector as any, \
    FilterCollector as filter
# noinspection PyUnresolvedReferences
from robot.collector.core import AsyncFnCollector as afn
# noinspection PyUnresolvedReferences
from robot.collector.core import AsyncForeachCollector as aforeach
# noinspection PyUnresolvedReferences
from robot.collector.core import ConstCollector as const
# noinspection PyUnresolvedReferences
from robot.collector.core import DefaultCollector as default
# noinspection PyUnresolvedReferences
from robot.collector.core import DelayCollector as delay
# noinspection PyUnresolvedReferences
# noinspection PyUnresolvedReferences
from robot.collector.core import FnCollector as fn
# noinspection PyUnresolvedReferences
from robot.collector.core import ForeachCollector as foreach
# noinspection PyUnresolvedReferences
from robot.collector.core import PipeCollector as pipe
# noinspection PyUnresolvedReferences
from robot.collector.core import TapCollector as tap
# noinspection PyUnresolvedReferences
# noinspection PyUnresolvedReferences
from robot.collector.core import TupleCollector as tuple
# noinspection PyUnresolvedReferences
from robot.collector.css import CssCollector as css
# noinspection PyUnresolvedReferences
from robot.collector.http import GetCollector as get, \
    UrlCollector as url
# noinspection PyUnresolvedReferences
from robot.collector.io import DownloadCollector as download
# noinspection PyUnresolvedReferences
from robot.collector.json import JsonPathCollector as jsonpath
# noinspection PyUnresolvedReferences
from robot.collector.store import StoreCollector as store, \
    CsvCollector as csv, \
    DictCsvCollector as dict_csv
# noinspection PyUnresolvedReferences
from robot.collector.text import RegexCollector as regex
# noinspection PyUnresolvedReferences
from robot.collector.xml import XPathCollector as xpath, \
    AttrCollector as attr, \
    AsTextCollector as as_text, \
    TextCollector as text

noop = lambda: _core.NOOP_COLLECTOR

context = lambda: _core.CONTEXT

flat = lambda: _collections.FLAT_COLLECTOR

chain = lambda: _collections.CHAIN_COLLECTOR


def dict(*args: Collector[X, Dict[str, Any]], **kwargs: Collector[X, Any]) -> Collector[X, Dict[str, Any]]:
    return _core.DictCollector(
        args,
        kwargs
    )


def obj(*args: Collector[X, Dict[str, Any]], **kwargs: Collector[X, Any]) -> Collector[X, _object.Object]:
    dict_collector = _core.DictCollector(
        args,
        kwargs
    )
    return _object.ObjectCollector(dict_collector)



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from logging import Logger

from robot.api import Collector, Context, XmlNode, Tuple

__logger__ = logging.getLogger(__name__)


@dataclass()
class GetCollector(Collector[str, XmlNode]):
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: str) -> Tuple[Context, XmlNode]:
        sub_context, sub_item = await context.http_get(item)
        return sub_context, sub_item


@dataclass()
class UrlCollector(Collector[str, str]):
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: str) -> Tuple[Context, str]:
        return context, context.resolve_url(item)



import logging
import re
from dataclasses import dataclass, field
from logging import Logger
from typing import Union, Sequence, Tuple, Dict

from robot.api import Collector, Context

__logger__ = logging.getLogger(__name__)


@dataclass(init=False)
class RegexCollector(Collector[str, str]):
    logger: Logger = field(default=__logger__, compare=False)

    def __init__(self, regex, logger=__logger__):
        if isinstance(regex, (str,)):
            regex = re.compile(regex)
        self.regex = regex
        self.logger = logger

    async def __call__(self, context: Context, item: str) -> Tuple[Context, Union[str, Sequence[str], Dict]]:
        match = self.regex.search(item)
        if not match:
            return context, None
        group_dict = match.groupdict()
        if group_dict:
            return context, group_dict
        groups = match.groups()
        if len(groups) > 1:
            return context, groups
        elif groups:
            return context, groups[0]
        return context, match.group(0)



import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import Iterable, Tuple

from robot.api import Collector, XmlNode, Context

__logger__ = logging.getLogger(__name__)


@dataclass()
class XPathCollector(Collector[XmlNode, XmlNode]):
    xpath: str
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: XmlNode) -> Tuple[Context, XmlNode]:
        return context, item.find_by_xpath(self.xpath)


@dataclass()
class AttrCollector(Collector[XmlNode, Iterable[str]]):
    attr: str
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: XmlNode) -> Tuple[Context, Iterable[str]]:
        return context, [
            value
            for value in item.attr(self.attr)
        ]


@dataclass()
class AsTextCollector(Collector[XmlNode, str]):
    prefix: str = ''
    suffix: str = ''
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: XmlNode) -> Tuple[Context, str]:
        if item.is_emtpy():
            return context, None
        return context, self.prefix + item.as_text() + self.suffix


@dataclass()
class TextCollector(Collector[XmlNode, Iterable[str]]):
    prefix: str = ''
    suffix: str = ''
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: XmlNode) -> Tuple[Context, Iterable[str]]:
        return context, [
            self.prefix + value + self.suffix
            for value in item.text()
        ]



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import Iterator, Iterable

from pyquery import PyQuery

from robot.api import XmlEngine, XmlNode

__logger__ = logging.getLogger(__name__)


@dataclass()
class PyQueryNodeAdapter(XmlNode):
    engine: PyQueryAdapter = field(compare=False)
    content: PyQuery
    logger: Logger = field(default=__logger__, compare=False)

    def __init__(self, engine: PyQueryAdapter, content: PyQuery):
        self.engine = engine
        self.content = content

    def __iter__(self) -> Iterator[PyQueryNodeAdapter]:
        return self._map(self.content)

    def _map(self, pyquery: PyQuery) -> Iterator[PyQueryNodeAdapter]:
        for item in pyquery:
            yield PyQueryNodeAdapter(self.engine, self.engine.pyquery(item))

    def find_by_css(self, css: str) -> XmlNode:
        return PyQueryNodeAdapter(self.engine, self.content.find(css))

    def find_by_xpath(self, xpath: str) -> XmlNode:
        return PyQueryNodeAdapter(self.engine, self.engine.pyquery(self.content.root.xpath(xpath)))

    def as_text(self) -> str:
        try:
            return self.content.text()
        except AttributeError:
            return ''.join([str(it) for it in self.content])

    def text(self) -> Iterable[str]:
        for item in self:
            yield item.content.text()

    def attr(self, attr: str) -> Iterable[str]:
        for item in self:
            yield item.content.attr(attr)

    def is_emtpy(self) -> bool:
        return len(self.content) == 0

    def __repr__(self):
        return f'{self.__class__.__name__}{{ {self.content} }}'

    def __str__(self):
        return f'{self.__class__.__name__}{{ {self.content} }}'


@dataclass()
class PyQueryAdapter(XmlEngine):
    pyquery: PyQuery = field(default=PyQuery, compare=False)
    logger: Logger = field(default=__logger__, compare=False)

    def __call__(self, raw_xml: str) -> PyQueryNodeAdapter:
        return PyQueryNodeAdapter(self, self.pyquery(raw_xml))



import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import Tuple

from robot.api import Collector, XmlNode, Context

__logger__ = logging.getLogger(__name__)


@dataclass()
class CssCollector(Collector[XmlNode, XmlNode]):
    css_selector: str
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: XmlNode) -> Tuple[Context, XmlNode]:
        return context, item.find_by_css(self.css_selector)



from __future__ import annotations

from typing import Iterable, Iterator, Tuple
from typing import TypeVar, Generic

X = TypeVar('X')
Y = TypeVar('Y')


class HttpEngine(object):

    def session(self) -> HttpSession:
        raise NotImplementedError()


class HttpSession(object):

    async def download(self, url: str, filename: str):
        raise NotImplementedError()

    async def get(self, url):
        raise NotImplementedError()

    async def __aenter__(self):
        raise NotImplementedError()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError()

    async def close(self):
        raise NotImplementedError()


class XmlNode(Iterable['XmlNode']):

    def __iter__(self) -> Iterator[XmlNode]:
        raise NotImplementedError()

    def find_by_css(self, css: str) -> XmlNode:
        raise NotImplementedError()

    def find_by_xpath(self, xpath: str) -> XmlNode:
        raise NotImplementedError()

    def as_text(self) -> str:
        raise NotImplementedError()

    def attr(self, attr) -> Iterable[str]:
        raise NotImplementedError()

    def is_emtpy(self) -> bool:
        raise NotImplementedError()

    def text(self) -> Iterable[str]:
        raise NotImplementedError()


class XmlEngine(object):

    def __call__(self, raw_xml: str) -> XmlNode:
        raise NotImplementedError()


class Context(object):
    xml_engine: XmlEngine
    http_engine: HttpEngine

    async def close(self):
        raise NotImplementedError()

    def resolve_url(self, url: str) -> str:
        raise NotImplementedError()

    async def download(self, url: str, filename: str):
        raise NotImplementedError()

    async def http_get(self, url) -> Tuple[Context, XmlNode]:
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()


class Collector(Generic[X, Y]):

    async def __call__(self, context: Context, item: X) -> Tuple[Context, Y]:
        raise NotImplementedError()


class Robot(object):

    def __enter__(self) -> 'Robot':
        raise NotImplementedError()

    def __exit__(self, exc_type, exc_val, exc_tb):
        raise NotImplementedError()

    async def run(self, collector: Collector[None, Y]) -> Y:
        raise NotImplementedError()

    def sync_run(self, collector: Collector[None, Y]) -> Y:
        raise NotImplementedError()



from __future__ import annotations

import logging
from dataclasses import dataclass, field
from logging import Logger
from typing import Dict, Callable, Any, Tuple

from robot.api import Collector, X, Y, Context
from robot.collector.core import DictCollector

__logger__ = logging.getLogger(__name__)


class Object(object):

    @classmethod
    def from_dict(cls, data: Dict) -> Object:
        return cls(dict([
            (k, v if not isinstance(v, dict) else cls.from_dict(v))
            for k, v in data.items()
        ]))

    def __init__(self, attributes):
        self.__dict__.update(attributes)

    def __eq__(self, other):
        if other is None:
            return False
        return self.__dict__ == other.__dict__

    def __repr__(self):
        return f'{self.__class__.__name__}({self.__dict__})'

    def __getitem__(self, item):
        return self.__dict__.__getitem__(item)

    def __setitem__(self, key, value):
        return self.__dict__.__setitem__(key, value)

    def __iter__(self):
        return self.__dict__.__iter__()


@dataclass()
class ObjectCollector(Collector[X, Y]):
    dict_collector: DictCollector[X]
    cast: Callable[[Dict[str, Any]], Y] = field(default=Object.from_dict)
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: X) -> Tuple[Context, Y]:
        context, value = await self.dict_collector(context, item)
        result = self.cast(value)
        return context, result



import logging
from dataclasses import dataclass, field
from itertools import chain
from logging import Logger
from typing import Iterable, List, Callable, Tuple

from robot.api import Collector, X, Context

__logger__ = logging.getLogger(__name__)


@dataclass()
class FlatCollector(Collector[Iterable[Iterable[X]], List[X]]):
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Iterable[Iterable[X]]) -> Tuple[Context, List[X]]:
        return context, list(chain(*item))


FLAT_COLLECTOR = FlatCollector()


@dataclass()
class ChainCollector(Collector[Iterable[Iterable[X]], Iterable[X]]):
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Iterable[Iterable[X]]) -> Tuple[Context,Iterable[X]]:
        return context, chain(*item)


CHAIN_COLLECTOR = ChainCollector()


@dataclass()
class AnyCollector(Collector[Iterable[X], X]):
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Iterable[X]) -> Tuple[Context, X]:
        return context, next(iter(item))


@dataclass()
class FilterCollector(Collector):
    predicate: Callable[[X], bool]
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: Iterable) -> Tuple[Context, Iterable]:
        return context, [
            value
            for value in item
            if self.predicate(value)
        ]



import os

from setuptools import setup, find_packages

__version__ = '0.4.dev1610151867' 


def read(file_name):
    with open(os.path.join(os.path.dirname(__file__), file_name)) as f:
        return f.read()


def read_requirements(file_name):
    lines = filter(None, read(file_name).splitlines())
    for line in lines:
        if line.startswith('#'):
            continue
        if line.startswith('-'):
            continue
        if ';' in line:
            line, _ = line.split(';', 1)
        yield line


requirements = read_requirements('requirements.txt')
requirements_dev = read_requirements('requirements-dev.txt')

setup(
    name='ciag-robot',
    version=__version__,
    description='Python Library to Build Web Robots',
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    url='https://github.com/OpenCIAg/py-robot',
    download_url='https://github.com/OpenCIAg/py-robot/tree/%s/' % __version__,
    license='Apache License 2.0',
    author='Éttore Leandro Tognoli',
    author_email='ettore.tognoli@ciag.org.br',
    data_files=[
        'requirements.txt',
        'requirements-dev.txt',
        'LICENSE',
    ],
    packages=find_packages(
        './src/main/python/',
    ),
    package_dir={'': 'src/main/python'},
    include_package_data=True,
    keywords=['Robot', 'Web Crawler'],
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Framework :: AsyncIO',
        'Topic :: Internet :: WWW/HTTP :: Indexing/Search',
    ],
    install_requires=list(requirements),
    tests_require=list(requirements_dev),
)



import logging
import os
from dataclasses import field, dataclass
from logging import Logger
from urllib.parse import urlparse

from robot.api import Collector, Context
from typing import Tuple
__logger__ = logging.getLogger(__name__)


class FileNameCollector(Collector[str, str]):
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: str) -> Tuple[Context, str]:
        parsed_url = urlparse(item)
        path, filename = parsed_url.path.rsplit('/', 1)
        path = os.path.join('.', path[1:])
        os.makedirs(path, exist_ok=True)
        return context, os.path.join(path, filename)


FILE_NAME_COLLECTOR = FileNameCollector()


@dataclass()
class DownloadCollector(Collector[str, str]):
    filename: Collector[str, str] = field(default=FILE_NAME_COLLECTOR)
    logger: Logger = field(default=__logger__, compare=False)

    async def __call__(self, context: Context, item: str) -> Tuple[Context, str]:
        context, filename = await self.filename(context, item)
        await context.download(item, filename)
        return context, filename



from dataclasses import dataclass, field
import logging
from logging import Logger
from typing import Any, Sequence

from jsonpath_ng import parse as parse_jsonpath

from robot.api import Collector, Context
from typing import Tuple
__logger__ = logging.getLogger(__name__)


@dataclass(init=False)
class JsonPathCollector(Collector[Any, Any]):
    logger: Logger = field(default=__logger__, compare=False)

    def __init__(self, jsonpath, logger=__logger__):
        if isinstance(jsonpath, (str,)):
            jsonpath = parse_jsonpath(jsonpath)
        self.jsonpath = jsonpath
        self.logger = logger

    async def __call__(self, context: Context, item: Any) -> Tuple[Context, Sequence[Any]]:
        return context, [
            match.value
            for match in self.jsonpath.find(item)
        ]



from __future__ import annotations

import logging
from dataclasses import dataclass

from robot.collector.core import DictCollector, PipeCollector
from robot.collector.object import ObjectCollector, Object
from robot.collector.xml import AsTextCollector
from robot.collector.xml import XPathCollector

__logger__ = logging.getLogger(__name__)


@dataclass()
class OpenGraph(Object):
    title: str = None
    type: str = None
    locale: str = None
    description: str = None
    url: str = None
    site_name: str = None
    image: str = None
    audio: str = None
    video: str = None
    app_id: str = None


dict_opengraph = DictCollector(
    (),
    {
        'title': PipeCollector(
            XPathCollector('//meta[@property="og:title"]/@content'),
            AsTextCollector(),
        ),
        'type': PipeCollector(
            XPathCollector('//meta[@property="og:type"]/@content'),
            AsTextCollector(),
        ),
        'locale': PipeCollector(
            XPathCollector('//meta[@property="og:locale"]/@content'),
            AsTextCollector(),
        ),
        'description': PipeCollector(
            XPathCollector('//meta[@property="og:description"]/@content'),
            AsTextCollector(),
        ),
        'url': PipeCollector(
            XPathCollector('//meta[@property="og:url"]/@content'),
            AsTextCollector(),
        ),
        'site_name': PipeCollector(
            XPathCollector('//meta[@property="og:site_name"]/@content'),
            AsTextCollector(),
        ),
        'image': PipeCollector(
            XPathCollector('//meta[@property="og:image"]/@content'),
            AsTextCollector(),
        ),
        'audio': PipeCollector(
            XPathCollector('//meta[@property="og:audio"]/@content'),
            AsTextCollector(),
        ),
        'video': PipeCollector(
            XPathCollector('//meta[@property="og:video"]/@content'),
            AsTextCollector(),
        ),
        'app_id': PipeCollector(
            XPathCollector('//meta[@property="og:app_id"]/@content'),
            AsTextCollector(),
        ),
    }
)


class OpenGraphCollector(ObjectCollector):

    def __init__(self):
        super().__init__(
            dict_opengraph,
            lambda kwargs: OpenGraph(**kwargs)
        )


OPENGRAPH_COLLECTOR = OpenGraphCollector()

obj_opengraph = lambda: OPENGRAPH_COLLECTOR
